import subprocess


def test_call_flake8(tmpdir):
    tmp = tmpdir.join('tmp.py')
    tmp.write('')
    output = subprocess.check_output(
        ['flake8', str(tmp)],
        stderr=subprocess.STDOUT,
    )
    assert output == b''


def test_call_flake8_version():
    output = subprocess.check_output(
        ['flake8', '--version'],
        stderr=subprocess.STDOUT,
    )
    assert b'flake8-commas-x: ' in output



import os
import sys
import tokenize

import pycodestyle
from flake8 import utils

from flake8_commas._base import Token, get_comma_errors

C813 = 'C813 missing trailing comma in Python 3'
C814 = 'C814 missing trailing comma in Python 2'
C815 = 'C815 missing trailing comma in Python 3.5+'
C816 = 'C816 missing trailing comma in Python 3.6+'
C818 = 'C818 trailing comma on bare tuple prohibited'


def get_tokens(filename):
    if filename == 'stdin':
        file_contents = utils.stdin_get_value().splitlines(True)
    else:
        file_contents = pycodestyle.readlines(filename)
    file_contents_iter = iter(file_contents)

    def file_contents_next():
        return next(file_contents_iter)

    for t in tokenize.generate_tokens(file_contents_next):
        yield Token(t)


def test_one_line_dict():
    filename = get_absolute_path('data/one_line_dict.py')
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_multiline_good_dict():
    filename = get_absolute_path('data/multiline_good_dict.py')
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_multiline_bad_dict():
    filename = get_absolute_path('data/multiline_bad_dict.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 14, 'line': 2, 'message': 'C812 missing trailing comma'},
    ]


def test_bad_list():
    filename = get_absolute_path('data/bad_list.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 5, 'line': 4, 'message': 'C812 missing trailing comma'},
        {'col': 5, 'line': 10, 'message': 'C812 missing trailing comma'},
        {'col': 5, 'line': 17, 'message': 'C812 missing trailing comma'},
    ]


def test_bad_function_call():
    filename = get_absolute_path('data/bad_function_call.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 17, 'line': 3, 'message': 'C812 missing trailing comma'},
    ]


def test_multiline_bad_function_def():
    fixture = 'data/multiline_bad_function_def.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 13, 'line': 9, 'message': 'C812 missing trailing comma'},
    ]


def test_bad_function_one_param():
    fixture = 'data/multiline_bad_function_one_param.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 13, 'line': 2, 'message': 'C812 missing trailing comma'},
        {'col': 9, 'line': 8, 'message': 'C812 missing trailing comma'},
    ]


def test_good_empty_comma_context():
    fixture = 'data/good_empty_comma_context.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_comma_good_dict():
    fixture = 'data/comment_good_dict.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_no_comma_required_list_comprehension():
    fixture = 'data/list_comprehension.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_no_comma_required_dict_comprehension():
    fixture = 'data/dict_comprehension.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_no_comma_required_multiline_if():
    fixture = 'data/multiline_if.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_no_comma_required_multiline_subscript():
    fixture = 'data/multiline_index_access.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 14, 'line': 27, 'message': 'C812 missing trailing comma'},
        {'col': 14, 'line': 34, 'message': 'C812 missing trailing comma'},
    ]


def test_comma_required_multiline_subscript_with_slice():
    fixture = 'data/multiline_subscript_slice.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 14, 'line': 5, 'message': 'C812 missing trailing comma'},
        {'col': 14, 'line': 33, 'message': 'C812 missing trailing comma'},
    ]


def test_comma_required_after_unpack_in_non_def_python_3_5():
    fixture = 'data/unpack.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 12, 'line': 4, 'message': C816},
        {'col': 9, 'line': 11, 'message': C816},
        {'col': 15, 'line': 19, 'message': C816},
        {'col': 12, 'line': 26, 'message': C815},
        {'col': 23, 'line': 32, 'message': C815},
        {'col': 14, 'line': 39, 'message': C816},
        {'col': 12, 'line': 46, 'message': C815},
        {'col': 12, 'line': 50, 'message': C815},
        {'col': 9, 'line': 58, 'message': C815},
        {'col': 9, 'line': 62, 'message': C815},
        {'col': 9, 'line': 68, 'message': C816},
        {'col': 12, 'line': 75, 'message': C816},
        {'col': 14, 'line': 83, 'message': C816},
        {'col': 19, 'line': 112, 'message': C815},
    ]


def test_no_comma_required_in_parenth_form():
    fixture = 'data/parenth_form.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_comma_required_in_argument_list():
    fixture = 'data/callable_before_parenth_form.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 7, 'line': 7, 'message': 'C812 missing trailing comma'},
        {'col': 7, 'line': 15, 'message': 'C812 missing trailing comma'},
        {'col': 7, 'line': 23, 'message': 'C812 missing trailing comma'},
    ]


def test_comma_required_even_if_you_use_or():
    fixture = 'data/multiline_bad_or_dict.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 14, 'line': 3, 'message': 'C812 missing trailing comma'},
    ]


def test_comma_not_required_in_multiline_case():
    fixture = 'data/multiline_case.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_comma_required_in_multiline_case_nested():
    fixture = 'data/multiline_case_bad.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 20, 'line': 5, 'message': 'C812 missing trailing comma'},
    ]


def test_comma_required_in_f_string_call():
    fixture = 'data/f_strings.py'
    filename = get_absolute_path(fixture)

    expected = (
        [
            {
                'col': 12,
                'line': 7,
                'message': 'C819 trailing comma prohibited',
            },
            {
                'col': 6,
                'line': 11,
                'message': 'C812 missing trailing comma',
            },
        ]
        if sys.version_info >= (3, 12)
        else []
    )

    assert list(get_comma_errors(get_tokens(filename))) == expected


def test_comma_not_required_even_if_you_use_dict_for():
    fixture = 'data/multiline_good_single_keyed_for_dict.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_comma_not_required_in_parenth_form_string_splits():
    fixture = 'data/multiline_string.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_comma_not_required_in_comment_lines():
    fixture = 'data/good_list.py'
    filename = get_absolute_path(fixture)
    assert list(get_comma_errors(get_tokens(filename))) == []


base = 'data/keyword_before_parenth_form/'


def test_base():
    filename = get_absolute_path(base + 'base.py')
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_base_bad():
    filename = get_absolute_path(base + 'base_bad.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 5, 'line': 2, 'message': 'C812 missing trailing comma'},
        {'col': 10, 'line': 8, 'message': 'C812 missing trailing comma'},
        {'col': 7, 'line': 14, 'message': 'C812 missing trailing comma'},
        {'col': 11, 'line': 17, 'message': 'C812 missing trailing comma'},
        {'col': 7, 'line': 21, 'message': 'C812 missing trailing comma'},
        {'col': 11, 'line': 24, 'message': 'C812 missing trailing comma'},
    ]


def test_py2():
    filename = get_absolute_path(base + 'py2.py')
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_py2_bad():
    filename = get_absolute_path(base + 'py2_bad.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 7, 'line': 4, 'message': 'C812 missing trailing comma'},
        {'col': 11, 'line': 7, 'message': C814},
        {'col': 7, 'line': 12, 'message': 'C812 missing trailing comma'},
        {'col': 11, 'line': 15, 'message': C814},
        {'col': 7, 'line': 20, 'message': 'C812 missing trailing comma'},
        {'col': 11, 'line': 23, 'message': C814},
        {'col': 7, 'line': 28, 'message': 'C812 missing trailing comma'},
        {'col': 11, 'line': 31, 'message': C814},
    ]


def test_py2_3():
    filename = get_absolute_path(base + 'py2_3.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
        {'col': 9, 'line': 4, 'message': C813},
        {'col': 9, 'line': 8, 'message': C813},
    ]


def test_py3():
    filename = get_absolute_path(base + 'py3.py')
    assert list(get_comma_errors(get_tokens(filename))) == []


def test_prohibited():
    filename = get_absolute_path('data/prohibited.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
       {'col': 21, 'line': 1, 'message': 'C819 trailing comma prohibited'},
       {'col': 13, 'line': 3, 'message': 'C819 trailing comma prohibited'},
       {'col': 18, 'line': 5, 'message': 'C819 trailing comma prohibited'},
       {'col': 6, 'line': 10, 'message': 'C819 trailing comma prohibited'},
       {'col': 21, 'line': 12, 'message': 'C819 trailing comma prohibited'},
       {'col': 13, 'line': 14, 'message': 'C819 trailing comma prohibited'},
       {'col': 18, 'line': 16, 'message': 'C819 trailing comma prohibited'},
       {'col': 6, 'line': 21, 'message': 'C819 trailing comma prohibited'},
       {'col': 10, 'line': 27, 'message': 'C819 trailing comma prohibited'},
       {'col': 9, 'line': 29, 'message': 'C819 trailing comma prohibited'},
    ]


def test_bare():
    # Tests inspired by flake8_tuple https://git.io/vxstN
    filename = get_absolute_path('data/bare.py')
    assert list(get_comma_errors(get_tokens(filename))) == [
       {'col': 8, 'line': 7, 'message': C818},
       {'col': 19, 'line': 9, 'message': C818},
       {'col': 8, 'line': 16, 'message': C818},
       {'col': 10, 'line': 20, 'message': C818},
       {'col': 32, 'line': 27, 'message': C818},
       {'col': 26, 'line': 29, 'message': C818},
       {'col': 17, 'line': 32, 'message': C818},
    ]


def get_absolute_path(filepath):
    return os.path.join(os.path.dirname(__file__), filepath)



from __future__ import annotations

import collections
import importlib.metadata
import token as mod_token
import tokenize
from typing import Iterator, NamedTuple, TypedDict

try:
    __version__ = importlib.metadata.version('flake8-commas-x')
except importlib.metadata.PackageNotFoundError:
    __version__ = 'unknown'


class ErrorDict(TypedDict):
    message: str
    line: int
    col: int


# A parenthesized expression list yields whatever that expression list
# yields: if the list contains at least one comma, it yields a tuple;
# otherwise, it yields the single expression that makes up the expression
# list.

PYTHON_2_KWDS = {
    'and', 'as', 'assert', 'break', 'class', 'continue', 'def', 'del', 'elif',
    'else', 'except', 'exec', 'finally', 'for', 'from', 'global', 'if',
    'import', 'in', 'is', 'lambda', 'not', 'or', 'pass', 'print', 'raise',
    'return', 'try', 'while', 'with', 'yield', 'case', 'match',
}

PYTHON_3_KWDS = {
    'False', 'None', 'True', 'and', 'as', 'assert', 'break', 'class',
    'continue', 'def', 'del', 'elif', 'else', 'except', 'finally', 'for',
    'from', 'global', 'if', 'import', 'in', 'is', 'lambda', 'nonlocal', 'not',
    'or', 'pass', 'raise', 'return', 'try', 'while', 'with', 'yield', 'case',
    'match',
}

KWD_LIKE_FUNCTION = {'import'}

ALL_KWDS = (PYTHON_2_KWDS & PYTHON_3_KWDS) - KWD_LIKE_FUNCTION
NOT_PYTHON_2_KWDS = (PYTHON_3_KWDS - PYTHON_2_KWDS) - KWD_LIKE_FUNCTION
NOT_PYTHON_3_KWDS = (PYTHON_2_KWDS - PYTHON_3_KWDS) - KWD_LIKE_FUNCTION


class FalsyObj(object):
    def __bool__(self):
        return False

    __nonzero__ = __bool__


TUPLE_OR_PARENTH_FORM = FalsyObj()
SUBSCRIPT = FalsyObj()
TUPLE_ISH = {SUBSCRIPT, TUPLE_OR_PARENTH_FORM}
LAMBDA_EXPR = FalsyObj()


class SimpleToken(object):
    def __init__(self, token: Token | None, type: str | None) -> None:
        self.token = token
        self.type = type

    def __repr__(self) -> str:
        return f'SimpleToken({self.token!r}, {self.type!r})'


class Context(NamedTuple):
    comma: bool | FalsyObj | str = False
    unpack: bool = False
    n: int = 0


NEW_LINE = 'new-line'
COMMA = ','
OPENING_BRACKET = '('
OPENING_SQUARE_BRACKET = '['
SOME_CLOSING = 'some-closing'
SOME_OPENING = 'some-opening'
OPENING = {SOME_OPENING, OPENING_BRACKET, OPENING_SQUARE_BRACKET}
CLOSING = {SOME_CLOSING}
BACK_TICK = '`'
CLOSE_ATOM = CLOSING | {BACK_TICK}
FOR = 'for'
NAMED = 'named'
PY2_ONLY_ERROR = 'py2-only-error'
PY3K_ONLY_ERROR = 'py3-only-error'
DEF = 'def'
FUNCTION_DEF = 'function-def'
FUNCTION = {NAMED, PY2_ONLY_ERROR, PY3K_ONLY_ERROR, FUNCTION_DEF}
UNPACK = '* or **'
ASSERT = 'assert'
LAMBDA = 'lambda'
COLON = ':'
NONE = SimpleToken(token=None, type=None)


def get_type(token: Token) -> str | None:
    type = token.type
    if type == tokenize.NL:
        return NEW_LINE

    string = token.string
    if type == tokenize.NAME and string == 'for':
        return FOR
    if type == tokenize.NAME and string == 'def':
        return DEF
    if type == tokenize.NAME and string == 'assert':
        return ASSERT
    if type == tokenize.NAME and string == 'lambda':
        return LAMBDA
    if type == mod_token.NAME and string not in ALL_KWDS:
        if string in NOT_PYTHON_2_KWDS:
            return PY2_ONLY_ERROR
        if string in NOT_PYTHON_3_KWDS:
            return PY3K_ONLY_ERROR
        return NAMED
    if type == tokenize.OP:
        if string in {'**', '*'}:
            return UNPACK
        if string == ',':
            return COMMA
        if string == '(':
            return OPENING_BRACKET
        if string == '[':
            return OPENING_SQUARE_BRACKET
        if string in {'[', '{'}:
            return SOME_OPENING
        if string in {']', ')', '}'}:
            return SOME_CLOSING
        if string == '`':
            return BACK_TICK
        if string == ':':
            return COLON
    return None


def simple_tokens(tokens: Iterator[Token]) -> Iterator[SimpleToken]:
    tokens = (t for t in tokens if t.type != tokenize.COMMENT)

    token = next(tokens)
    previous_token = SimpleToken(token=token, type=get_type(token))
    for token in tokens:
        next_token = SimpleToken(token=token, type=get_type(token))
        if previous_token.type == NEW_LINE and next_token.type == NEW_LINE:
            continue
        yield previous_token
        previous_token = next_token


ERRORS = {
    True: ('C812', 'missing trailing comma'),
    SUBSCRIPT: ('C812', 'missing trailing comma'),
    TUPLE_OR_PARENTH_FORM: ('C812', 'missing trailing comma'),
    FUNCTION_DEF: ('C812', 'missing trailing comma'),
    PY3K_ONLY_ERROR: ('C813', 'missing trailing comma in Python 3'),
    PY2_ONLY_ERROR: ('C814', 'missing trailing comma in Python 2'),
    'py35': ('C815', 'missing trailing comma in Python 3.5+'),
    'py36': ('C816', 'missing trailing comma in Python 3.6+'),
}


def process_parentheses(
    token: SimpleToken,
    prev_1: SimpleToken,
    prev_2: SimpleToken,
) -> list[Context]:
    previous_token = prev_1

    if token.type == OPENING_BRACKET:
        is_function = (
            previous_token and
            (
                (previous_token.type in CLOSE_ATOM) or
                (
                    previous_token.type in FUNCTION
                )
            )
        )
        if is_function:
            if prev_2.type == DEF:
                return [Context(FUNCTION_DEF)]
            tk_string = previous_token.type
            if tk_string == PY2_ONLY_ERROR:
                return [Context(PY2_ONLY_ERROR)]
            if tk_string == PY3K_ONLY_ERROR:
                return [Context(PY3K_ONLY_ERROR)]
        else:
            return [Context(TUPLE_OR_PARENTH_FORM)]

    if token.type == OPENING_SQUARE_BRACKET:
        is_index_access = (
            previous_token and
            (
                (previous_token.type in CLOSING) or
                (
                    previous_token.type == NAMED
                )
            )
        )
        if is_index_access:
            return [Context(SUBSCRIPT)]

    return [Context(True)]


def get_comma_errors(tokens_iter: Iterator[Token]) -> Iterator[ErrorDict]:
    tokens = simple_tokens(tokens_iter)

    stack = [Context()]

    window = collections.deque([NONE, NONE], maxlen=3)

    for token in tokens:
        window.append(token)
        prev_2, prev_1, _ = window
        if token.type in OPENING:
            stack.extend(
                process_parentheses(token, prev_1, prev_2),
            )

        if token.type == LAMBDA:
            stack.append(Context(LAMBDA_EXPR))

        if token.type == FOR:
            stack[-1] = Context()

        if token.type == COMMA:
            stack[-1] = stack[-1]._replace(n=stack[-1].n + 1)

        if token.type == UNPACK:
            stack[-1] = stack[-1]._replace(unpack=True)

        comma_allowed = token.type in CLOSING and (
            stack[-1].comma or
            stack[-1].comma in TUPLE_ISH and stack[-1].n >= 1
        )

        comma_prohibited = prev_1.type == COMMA and (
            (
                comma_allowed and
                (stack[-1].comma not in TUPLE_ISH or stack[-1].n > 1)
            ) or stack[-1].comma == LAMBDA_EXPR and token.type == COLON
        )
        if comma_prohibited:
            end_row, end_col = prev_1.token.end
            yield {
                'message': 'C819 trailing comma prohibited',
                'line': end_row,
                'col': end_col,
            }

        bare_comma_prohibited = (
            token.token.type == tokenize.NEWLINE and
            prev_1.type == COMMA
        )

        if bare_comma_prohibited:
            end_row, end_col = prev_1.token.end
            yield {
                'message': 'C818 trailing comma on bare tuple prohibited',
                'line': end_row,
                'col': end_col,
            }

        comma_required = (
            comma_allowed and
            prev_1.type == NEW_LINE and
            prev_2.type != COMMA and
            prev_2.type not in OPENING
        )
        if comma_required:
            end_row, end_col = prev_2.token.end
            if (stack[-1].unpack and stack[-1].comma == FUNCTION_DEF):
                errors = ERRORS['py36']
            elif (stack[-1].unpack):
                errors = ERRORS['py35']
            else:
                errors = ERRORS[stack[-1].comma]
            yield {
                'message': '%s %s' % errors,
                'line': end_row,
                'col': end_col,
            }

        pop_stack = (
            token.type in CLOSING or
            (token.type == COLON and stack[-1].comma == LAMBDA_EXPR)
        )
        if pop_stack:
            stack.pop()


class CommaChecker(object):
    name = __name__
    version = __version__

    def __init__(self, tree, filename='(none)', file_tokens=None):
        fn = 'stdin' if filename in ('stdin', '-', None) else filename
        self.filename = fn
        self.tokens = file_tokens

    def run(self):
        file_tokens = self.tokens
        tokens = (Token(t) for t in file_tokens)

        for error in get_comma_errors(tokens):
            yield (
                error.get('line'),
                error.get('col'),
                error.get('message'),
                type(self),
            )


class Token:
    """Python 2 and 3 compatible token"""

    def __init__(self, token: tokenize.TokenInfo) -> None:
        self.token = token

    @property
    def type(self):
        return self.token[0]

    @property
    def string(self):
        return self.token[1]

    @property
    def start(self):
        return self.token[2]

    @property
    def start_row(self):
        return self.start[0]

    @property
    def start_col(self):
        return self.start[1]

    @property
    def end(self):
        return self.token[3]

    @property
    def end_row(self):
        return self.end[0]

    @property
    def end_col(self):
        return self.end[1]

    def __repr__(self) -> str:
        return f'Token({self.token!r})'



import io
import os

from setuptools import setup

__dir__ = os.path.dirname(__file__)


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


readme = read('README.rst')
history = read('CHANGES.rst').replace('.. :changelog:', '')


setup(
    name='flake8-commas-x',
    author="Peter Law",
    author_email="PeterJCLaw@gmail.com",
    maintainer="Peter Law",
    maintainer_email="PeterJCLaw@gmail.com",
    version='3.0.0',
    install_requires=['flake8>=5'],
    python_requires='>=3.8',
    url='https://github.com/PeterJCLaw/flake8-commas/',
    long_description=readme + '\n\n' + history,
    description='Flake8 lint for trailing commas.',
    packages=['flake8_commas'],
    test_suite='test',
    include_package_data=True,
    entry_points={
        'flake8.extension': [
            'C81 = flake8_commas:CommaChecker',
        ],
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
        'Framework :: Flake8',
    ],
)
